#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/ioctl.h>

#include "mctd.h"
#include "aiot-modem-control-help.h"

#define UNUSED(x) ((void)(x))

typedef struct {
    unsigned int id;
    char* str;
} MCTD_CMD_STR_TO_ID_T;

const MCTD_CMD_STR_TO_ID_T cmd_str_to_id_table[] = {
    { MCTD_CMD_SET_APN,            MCTD_CMD_STR_SET_APN            },
    { MCTD_CMD_DIAL,               MCTD_CMD_STR_DIAL               },
    { MCTD_CMD_HANGUP,             MCTD_CMD_STR_HANGUP             },
    { MCTD_CMD_GET_PHONE_NUMBER,   MCTD_CMD_STR_GET_PHONE_NUMBER   },
    { MCTD_CMD_GET_SIGNAL_QUALITY, MCTD_CMD_STR_GET_SIGNAL_QUALITY },
    { MCTD_CMD_WWAN_FORCE_RESTART, MCTD_CMD_STR_WWAN_FORCE_RESTART },
    { MCTD_CMD_POWERON,            MCTD_CMD_STR_POWERON            },
    { MCTD_CMD_POWEROFF,           MCTD_CMD_STR_POWEROFF           },
    { MCTD_CMD_SET_PSM,            MCTD_CMD_STR_SET_PSM            },
    { MCTD_CMD_SET_EDRX,           MCTD_CMD_STR_SET_EDRX           },
    { MCTD_CMD_SET_SUSPEND,        MCTD_CMD_STR_SET_SUSPEND        },
    { MCTD_CMD_SET_SLEEP,          MCTD_CMD_STR_SET_SLEEP          },
    { MCTD_CMD_ACTIVATE,           MCTD_CMD_STR_ACTIVATE           },
    { MCTD_CMD_DEACTIVATE,         MCTD_CMD_STR_DEACTIVATE         },
    { MCTD_CMD_SEND_AT,            MCTD_CMD_STR_SEND_AT            },
    { MCTD_CMD_SEND_AT_ECHO,       MCTD_CMD_STR_SEND_AT_ECHO       },
    { MCTD_CMD_SET_SIM_SLOT,       MCTD_CMD_STR_SET_SIM_SLOT       },
    { MCTD_CMD_UPDATE_DATETIME,    MCTD_CMD_STR_UPDATE_DATETIME    },
    { MCTD_CMD_GET_DATETIME,       MCTD_CMD_STR_GET_DATETIME       },
    { MCTD_CMD_POWEROFF_AND_FIN,   MCTD_CMD_STR_POWEROFF_AND_FIN   },
    { MCTD_CMD_FIN,                MCTD_CMD_STR_FIN                },
    { MCTD_CMD_STATUS,             MCTD_CMD_STR_STATUS             },
    { MCTD_CMD_LIST_SMS,           MCTD_CMD_STR_LIST_SMS           },
    { MCTD_CMD_SHOW_SMS,           MCTD_CMD_STR_SHOW_SMS           },
    { MCTD_CMD_CREATE_SMS,         MCTD_CMD_STR_CREATE_SMS         },
    { MCTD_CMD_SEND_SMS,           MCTD_CMD_STR_SEND_SMS           },
    { MCTD_CMD_DELETE_SMS,         MCTD_CMD_STR_DELETE_SMS         },
    { MCTD_CMD_SET_SMS_STORAGE,    MCTD_CMD_STR_SET_SMS_STORAGE    },
};

static bool is_valid_command(const int argc, char *argv[], unsigned int *cmd_id)
{
    unsigned int i;
    unsigned int array_size =
                    sizeof(cmd_str_to_id_table)/sizeof(MCTD_CMD_STR_TO_ID_T);
    if (argc < 2) {
        return false;
    }

    /* check commands */
    for(i = 0; i < array_size; i++) {
        if (strcmp(argv[1], cmd_str_to_id_table[i].str) == 0) {
            *cmd_id = cmd_str_to_id_table[i].id;
            return true;
        }
    }
    return false;
}

static int send_request(const int fd, unsigned int cmd_id, const char *param)
{
    MCTD_REQUEST_T request;

    /* send header */
    if (write(fd, MCTD_HEADER, strlen(MCTD_HEADER)) == -1) {
        perror("write header");
        return MCTD_RESULT_ERR_SOCKET;
    }

    /* write fix parameter */
    mctd_u32tou8(htonl(cmd_id), request.cmd_id);

    if (param)
        mctd_u32tou8(htonl(strlen(param)), request.param_len);
    else
        mctd_u32tou8(0, request.param_len);

    if (write(fd, &request, sizeof(MCTD_REQUEST_T)) == -1) {
        perror("write fix parameter.");
        return MCTD_RESULT_ERR_SOCKET;
    }

    /* write parameter if exist */
    if (param[0] == '\0')
        return MCTD_RESULT_OK;

    if (write(fd, param, strlen(param)) == -1) {
        perror("write parameter.");
        return MCTD_RESULT_ERR_SOCKET;
    }

    return MCTD_RESULT_OK;
}

int receive_response(int fd, unsigned int *cmd_id, unsigned int *param_len,
                                                                char** param)
{
    char buf[32];
    MCTD_RESPONSE_T response;
    int result;

    /* read header */
    if (mctd_receive_data(fd, buf, strlen(MCTD_HEADER), 90) != MCTD_RESULT_OK) {
        printf("read header failed.\n");
        return MCTD_RESULT_ERR_SOCKET;
    }

    /* check header */
    if (!mctd_is_valid_header(buf)) {
        printf("check header failed.\n");
        return MCTD_RESULT_ERR_SOCKET;
    }

    /* read fix parameter */
    if (mctd_receive_data(fd, buf, sizeof(MCTD_RESPONSE_T), 3)
                                                         != MCTD_RESULT_OK) {
        printf("read fix parameter failed.\n");
        return MCTD_RESULT_ERR_SOCKET;
    }

    memcpy(&response, buf, sizeof(MCTD_RESPONSE_T));

    *cmd_id = ntohl(mctd_u8tou32(response.cmd_id));
    result = (int)(ntohl(mctd_u8tou32(response.result)));
    *param_len = ntohl(mctd_u8tou32(response.param_len));

    /* read parameter if exist */
    if (*param_len == 0) {
        *param = NULL;
        return result;
    }

    *param = malloc(*param_len+1);
    if (!(*param)) {
        perror("malloc");
        return MCTD_RESULT_ERR_MEMORY;
    }

    memset(*param, 0, *param_len+1);

    if (mctd_receive_data(fd, *param, *param_len, 3) != MCTD_RESULT_OK) {
        printf("read parameter failed.\n");
        free(*param);
        *param = NULL;
        return MCTD_RESULT_ERR_SOCKET;
    }

    return result;
}

static int parse_param_set_apn(int argc, char *argv[])
{
    UNUSED(argv);

    if (argc != 3)
        return MCTD_RESULT_ERR_PARAM;

    return MCTD_RESULT_OK;
}

static int parse_param_set_psm_tau(const char *param)
{
    char str[5];
    int len;
    long num;

    len = strlen(param);
    if (len < 2 || len > 5)
        return MCTD_RESULT_ERR_PARAM;

    memset(str, 0, 5);
    strcpy(str, param);
    if (str[len-1] == 's') {
        str[len-1] = '\0';
        num = strtol(str, NULL, 10);
        if (num >= MCTD_PSM_2SEC*MCTD_PSM_MIN_BIT &&
            ((!(num % MCTD_PSM_2SEC) &&
              (num <= MCTD_PSM_2SEC*MCTD_PSM_MAX_BIT)) ||
             (!(num % MCTD_PSM_30SEC) &&
              (num <= MCTD_PSM_30SEC*MCTD_PSM_MAX_BIT))))
            return MCTD_RESULT_OK;
    } else if ((str[len-1] == 'm') || (str[len-1] == 'h')) {
        /* min and hour is the same logic. */
        str[len-1] = '\0';
        num = strtol(str, NULL, 10);
        if (num >= MCTD_PSM_1MIN &&
            ((num <= MCTD_PSM_1MIN*MCTD_PSM_MAX_BIT) ||
             (!(num % MCTD_PSM_10MIN) &&
              (num <= MCTD_PSM_10MIN*MCTD_PSM_MAX_BIT))))
            return MCTD_RESULT_OK;
    }
    return MCTD_RESULT_ERR_PARAM;
}

static int parse_param_set_psm_act_time(const char *param)
{
    char str[5];
    int len;
    long num;

    len = strlen(param);
    if (len < 2 || len > 5)
        return MCTD_RESULT_ERR_PARAM;

    memset(str, 0, 5);
    strcpy(str, param);
    if (str[len-1] == 's') {
        str[len-1] = '\0';
        num = strtol(str, NULL, 10);
        if (num >= MCTD_PSM_2SEC &&
            (!(num % MCTD_PSM_2SEC) &&
             (num <= MCTD_PSM_2SEC*MCTD_PSM_MAX_BIT)))
            return MCTD_RESULT_OK;
    } else if (str[len-1] == 'm') {
        str[len-1] = '\0';
        num = strtol(str, NULL, 10);
        if (num >= MCTD_PSM_1MIN &&
            ((num <= MCTD_PSM_1MIN*MCTD_PSM_MAX_BIT) ||
             (!(num % MCTD_PSM_DECIHOURS) &&
              (num <= MCTD_PSM_DECIHOURS*MCTD_PSM_MAX_BIT))))
            return MCTD_RESULT_OK;
    }
    return MCTD_RESULT_ERR_PARAM;
}

static int parse_param_set_psm(int argc, char *argv[])
{
    if (argc < 3)
        return MCTD_RESULT_ERR_PARAM;

    if ((strcmp(MCTD_STR_DISABLE, argv[2]) == 0) ||
        (strcmp(MCTD_STR_DEFAULT, argv[2]) == 0))
        return MCTD_RESULT_OK;

    if (strcmp(MCTD_STR_ENABLE, argv[2]) != 0)
        return MCTD_RESULT_ERR_PARAM;

    if (argc < 5)
        return MCTD_RESULT_ERR_PARAM;

    if (parse_param_set_psm_tau(argv[3]) != MCTD_RESULT_OK)
        return MCTD_RESULT_ERR_PARAM;

    if (parse_param_set_psm_act_time(argv[4]) != MCTD_RESULT_OK)
        return MCTD_RESULT_ERR_PARAM;

    return MCTD_RESULT_OK;
}

const char* param_set_edrx_pcl[] = {
    "5.12", "10.24", "20.48", "40.96", "61.44", "81.92", "102.4", "122.88",
    "143.36", "163.84", "327.68", "655.36", "1310.72", "2621.44", NULL
};

const char* param_set_edrx_ptw[] = {
    "1.28", "2.56", "5.12", "6.40", "7.68", "8.96", "10.24", "11.52", "12.80",
    "14.08", "15.36", "16.64", "17.92", "19.20", "20.48", NULL
};

static int parse_param_set_edrx(int argc, char *argv[])
{
    int i = 0;
    bool found = false;

    if (argc < 3)
        return MCTD_RESULT_ERR_PARAM;

    if ((strcmp(MCTD_STR_DISABLE, argv[2]) == 0) ||
        (strcmp(MCTD_STR_DEFAULT, argv[2]) == 0))
        return MCTD_RESULT_OK;

    if (strcmp(MCTD_STR_ENABLE, argv[2]) != 0)
        return MCTD_RESULT_ERR_PARAM;


    if (argc < 5)
        return MCTD_RESULT_ERR_PARAM;

    while(param_set_edrx_pcl[i]) {
        if (strcmp(param_set_edrx_pcl[i], argv[3]) == 0) {
            found = true;
            break;
        }
        i++;
    }

    if (!found)
        return MCTD_RESULT_ERR_PARAM;

    i = 0;
    found = false;
    while(param_set_edrx_ptw[i]) {
        if (strcmp(param_set_edrx_ptw[i], argv[4]) == 0) {
            found = true;
            break;
        }
        i++;
    }

    return found ? MCTD_RESULT_OK : MCTD_RESULT_ERR_PARAM;
}

static int parse_param_set_suspend(int argc, char *argv[])
{
    if (argc != 3)
        return MCTD_RESULT_ERR_PARAM;

    if (strcmp(MCTD_STR_ENABLE, argv[2]) != 0 &&
        strcmp(MCTD_STR_DISABLE, argv[2]) != 0)
        return MCTD_RESULT_ERR_PARAM;

    return MCTD_RESULT_OK;
}

static int parse_param_set_sleep(int argc, char *argv[])
{
    long msec;
    if (argc < 3)
        return MCTD_RESULT_ERR_PARAM;

    if (strcmp(MCTD_STR_ENABLE, argv[2]) == 0) {
        if (argc < 4)
            return MCTD_RESULT_ERR_PARAM;

        msec = strtol(argv[3], NULL, 10);
        if (msec >= MCTD_SPOW_MIN_MSEC && msec <= MCTD_SPOW_MAX_MSEC)
            return MCTD_RESULT_OK;
    } else if (strcmp(MCTD_STR_DISABLE, argv[2]) == 0)
        return MCTD_RESULT_OK;

    return MCTD_RESULT_ERR_PARAM;
}

static int parse_param_send_at(int argc, char *argv[])
{
    UNUSED(argv);

    if (argc != 3)
        return MCTD_RESULT_ERR_PARAM;

    if (strlen(argv[2]) >= MCTD_AT_BUF_SIZE)
        return MCTD_RESULT_ERR_PARAM;

    return MCTD_RESULT_OK;
}

static int parse_param_set_sim_slot(int argc, char *argv[])
{
    UNUSED(argv);

    if (argc != 3)
        return MCTD_RESULT_ERR_PARAM;

    if ((strcmp(argv[2], "1") != 0) && (strcmp(argv[2], "2") != 0))
        return MCTD_RESULT_ERR_PARAM;

    return MCTD_RESULT_OK;
}

static int parse_param_update_datetime(int argc, char *argv[])
{
    UNUSED(argv);

    if (argc != 3)
        return MCTD_RESULT_ERR_PARAM;

    return MCTD_RESULT_OK;
}

typedef struct {
    unsigned int cmd_id;
    int (*func)(int, char **);
} CMD_PARSE_T;

const CMD_PARSE_T cmd_parse_table[] = {
    { MCTD_CMD_SET_APN,          parse_param_set_apn          },
    { MCTD_CMD_SET_PSM,          parse_param_set_psm          },
    { MCTD_CMD_SET_EDRX,         parse_param_set_edrx         },
    { MCTD_CMD_SET_SUSPEND,      parse_param_set_suspend      },
    { MCTD_CMD_SET_SLEEP,        parse_param_set_sleep        },
    { MCTD_CMD_SEND_AT,          parse_param_send_at          },
    { MCTD_CMD_SEND_AT_ECHO,     parse_param_send_at          },
    { MCTD_CMD_SET_SIM_SLOT,     parse_param_set_sim_slot     },
    { MCTD_CMD_UPDATE_DATETIME,  parse_param_update_datetime  },
};

static int parse_param(unsigned int cmd_id, int argc, char *argv[])
{
   int array_size = sizeof(cmd_parse_table)/sizeof(CMD_PARSE_T); 
   int i;

   for (i = 0; i < array_size; i++) {
        if (cmd_id == cmd_parse_table[i].cmd_id)
            return cmd_parse_table[i].func(argc, argv);
   }
   return MCTD_RESULT_OK;
}

typedef struct {
    int result;
    const char *str;
} RESULT_STR_T;

const RESULT_STR_T result_str_table[] = {
    { MCTD_RESULT_OK,                "OK"                },
    { MCTD_RESULT_ERR_CALLING,       "on Calling"        },
    { MCTD_RESULT_ERR_PROCESS,       "Command Failed"    },
    { MCTD_RESULT_ERR_SOCKET,        "Socket Error"      },
    { MCTD_RESULT_ERR_CMD_NOT_FOUND, "Command Not Found" },
    { MCTD_RESULT_ERR_MEMORY,        "Memory Error"      },
    { MCTD_RESULT_ERR_USING_SERIAL,  "Serial In Use"     },
    { MCTD_RESULT_ERR_SERIAL,        "Serial Error"      },
    { MCTD_RESULT_ERR_TIMEOUT,       "Timeout"           },
    { MCTD_RESULT_ERR_PARAM,         "Parameter Error"   },
};

static void put_result(const int result, const char *param)
{
    int i;
    int array_size = sizeof(result_str_table)/sizeof(RESULT_STR_T);

    if (result == MCTD_RESULT_OK) {
        if (!param)
            return;

        printf("%s\n", param);
        return;
    }

    for (i = 0; i < array_size; i++) {
        if (result == result_str_table[i].result) {
            printf("%s\n", result_str_table[i].str);
            return;
        }
    }
    printf("%s\n", "Command Failed");
}

int main(int argc, char *argv[])
{
    int fd;
    unsigned int cmd_id;
    unsigned int param_len = 0;
    unsigned int param_len_current;
    int result;
    int val;
    int i;
    char param[256];
    char *response = NULL;
    struct sockaddr_un sa = {0};

    if ((argc >= 2) && (strcmp(argv[1], "help") == 0)) {
        print_help();
        return 1;
    }

    if (!is_valid_command(argc, argv, &cmd_id)) {
        printf("unknown command\n");
        print_command();
        return 1;
    }

    if (parse_param(cmd_id, argc, argv) != MCTD_RESULT_OK) {
        put_result(MCTD_RESULT_ERR_PARAM, NULL);
        return 1;
    }

    memset(param, 0, 256);
    if (argc >= 3) {
        /* get parameter(s) */
        for(i = 2; i < argc; i++) {
            param_len_current = strlen(argv[i]);
            if ( (param_len_current + param_len) >= 255) {
                printf("parameter error\n");
                return 1;
            }
            strcpy(&param[param_len], argv[i]);
            param_len += param_len_current;
            if (i < (argc-1)) {
                param[param_len] = '\t';
                param_len++;
            }
        }
    }

    fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (fd == -1) {
        perror("socket");
        return 1;
    }

    /* set nonblock */
    val = 1;
    ioctl(fd, FIONBIO, &val);

    sa.sun_family = AF_UNIX;
    strcpy(sa.sun_path, SOCK_PATH);

    if (connect(fd, (struct sockaddr*) &sa, sizeof(struct sockaddr_un)) == -1) {
        perror("connect");
        goto bail;
    }

    if (send_request(fd, cmd_id, param) != MCTD_RESULT_OK) {
        goto bail;
    }

    result = receive_response(fd, &cmd_id, &param_len, &response);
    put_result(result, response);

    if (response)
        free(response);

    if (result != MCTD_RESULT_OK)
        goto bail;

    close(fd);
    return result;

bail:
    close(fd);
    return 1;
}
