#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/select.h>
#include <errno.h>
#include <sys/stat.h>

#include "mctd.h"

unsigned int mctd_u8tou32(unsigned char *u8)
{
    return u8[0] << 24 | u8[1] << 16 | u8[2] << 8 | u8[3];
}

void mctd_u32tou8(unsigned int u32, unsigned char *u8)
{
    u8[0] = (unsigned char)((u32 >> 24) & 0xff);
    u8[1] = (unsigned char)((u32 >> 16) & 0xff);
    u8[2] = (unsigned char)((u32 >>  8) & 0xff);
    u8[3] = (unsigned char)( u32        & 0xff);
}

bool mctd_is_valid_header(const char *header)
{
    return (strncmp(header, MCTD_HEADER, 4) == 0);
}

int mctd_wait_data_with_usec(int fd, int timeout_sec, int timeout_usec)
{
    struct timeval tv;
    fd_set readfds;
    int ret;

    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);

    tv.tv_sec = timeout_sec;
    tv.tv_usec = timeout_usec;

    ret = select(fd + 1, &readfds, NULL, NULL, &tv);

    if (ret == -1) {
        perror("select");
        return MCTD_RESULT_ERR_SOCKET;
    }

    if (!ret)
        return MCTD_RESULT_ERR_TIMEOUT;

    if (FD_ISSET(fd, &readfds))
        return MCTD_RESULT_OK;

    return MCTD_RESULT_ERR_SOCKET;
}

int mctd_wait_data(int fd, int timeout_sec)
{
    return mctd_wait_data_with_usec(fd, timeout_sec, 0);
}

int mctd_receive_data(int fd, char* buf, int desire_len, int retry)
{
    int done_len = 0;
    int recv_len;
    int i;
    int result;

    recv_len = read(fd, buf, desire_len);
    if (recv_len == desire_len)
            return MCTD_RESULT_OK;

    if (recv_len > 0)
        done_len = recv_len;

    for (i = 0; i < retry; i++) {
        result = mctd_wait_data(fd, 2);
        if (result == MCTD_RESULT_ERR_TIMEOUT)
            continue;

        if (result != MCTD_RESULT_OK)
            return MCTD_RESULT_ERR_SOCKET;

        recv_len = read(fd, &buf[done_len], desire_len - done_len);
        if ((recv_len == -1) && (errno != EAGAIN))
            return MCTD_RESULT_ERR_SOCKET;

        done_len += recv_len;
        if (recv_len == desire_len)
            return MCTD_RESULT_OK;
    }

    return MCTD_RESULT_ERR_TIMEOUT;
}

bool mctd_create_pid_file()
{
    FILE *fp;

    if (mctd_has_pid_file())
        return true;

    fp = fopen(PID_PATH, "w");
    if (!fp) {
        perror("fopen pid file:");
        return false;
    }

    fprintf(fp, "%d", getpid());
    fclose(fp);
    return true;
}

void mctd_remove_pid_file()
{
    remove(PID_PATH);
}

bool mctd_has_pid_file()
{
    return access(PID_PATH, F_OK) == 0;
}
